#------------------------------------------------------------
#    tree
#------------------------------------------------------------

module Creek
  #------------------------------------------------------------
  #    class Node
  #------------------------------------------------------------
  class Node
    def initialize(value0 = nil, parent0 = nil, older0 = nil, younger0 = nil)
      @value =  value0
      @parent = parent0
      @older = older0
      @younger = younger0
      @children = []
      return self
    end
    
    def append_child(value)
      if(!@children.empty?)
        nodetmp = Node.new(value, self, @children.last, nil)
        @children.last.set_younger(nodetmp)
        @children.push(nodetmp)
      else
        @children.push(Node.new(value, self, nil, nil))
      end
      return @children.last
    end
    
    def append_child_node(child_node)
      nodetmp = child_node
      nodetmp.set_parent(self)
      nodetmp.set_older(@children.last)
      nodetmp.set_younger(nil)
      if(!@children.empty?)
        @children.last.set_younger(nodetmp)
        @children.push(nodetmp)
      else
        @children.push(nodetmp)
      end
      return @children.last
    end
    
    def prepend_child(value)
      if(!@children.empty?)
        nodetmp = Node.new(value, self, nil, @children.first)
        @children.first.set_older(nodetmp)
        @children.unshift(nodetmp)
      else
        @children.push(Node.new(value, self, nil, nil))
      end
      return @children.first
    end
    
    def prepend_child_node(child_node)
      nodetmp = child_node
      nodetmp.set_parent(self)
      nodetmp.set_older(nil)
      nodetmp.set_younger(@children.first)
      if(!@children.empty?)
        @children.first.set_older(nodetmp)
        @children.unshift(nodetmp)
      else
        @children.push(nodetmp)
      end
      return @children.first
    end
    
    def insert_child(n, value)
      if(n < 0 || n > @children.size); return nil; end
      nodetmp = Node.new(value, self)
      if(n - 1 >= 0)
        nodetmp.set_older(@children[n - 1])
        @children[n - 1].set_younger(nodetmp)
      end
      if(n < @children.size)
        nodetmp.set_younger(@children[n])
        @children[n].set_older(nodetmp)
      end
      @children.insert(n, nodetmp)
      return nodetmp
    end
    
    def preorder_increment
      if(!@children.empty?)
        return @children.first
      elsif(@parent != nil)
        cur_node = self
        while(cur_node != nil)
          if(cur_node.younger != nil)
            then return cur_node.younger
          end
          cur_node = cur_node.parent
        end
        return self
      else
        return self
      end
    end
       
    def preorder_decrement
      if(@older != nil)
        cur_node = @older
        while(!cur_node.children.empty?)
          cur_node = cur_node.children.last
        end
        return cur_node
      elsif(@parent != nil)
        return @parent
      else
        return self
      end
    end
    
    def postorder_increment
      if(@younger != nil)
        cur_node = @younger
        while(!cur_node.children.empty?)
          cur_node = cur_node.children.first
        end
        return cur_node
      elsif(@parent != nil)
        return @parent
      else
        return self
      end
    end
       
    def postorder_decrement
      if(!@children.empty?)
        return @children.last
      else
        cur_node = self
        while(cur_node != nil)
          if(cur_node.older != nil)
            return cur_node.older
          end
          cur_node = cur_node.parent
        end
        return self
      end
    end
    
    def set_value(val); @value = val; end
    def set_parent(parent0); @parent = parent0; end
    def set_older(older0); @older = older0; end
    def set_younger(younger0); @younger = younger0; end
    attr_reader :value, :parent, :older, :younger, :children
  end #---- class Node
  
  #------------------------------------------------------------
  #    class Tree
  #------------------------------------------------------------
  class Tree
    def initialize
      @foot = Node.new(nil, nil, nil, nil)
      @root = @foot
      return self
    end
    
    def begin_preorder
      PreorderIterator.new(self, @root)
    end
    
    def end_preorder
      PreorderIterator.new(self, @foot)
    end
    
    def begin_postorder
      if(@root.equal?(@foot))
        return PostorderIterator.new(self, @foot)
      else
        tmp = @root
        while(!tmp.children.empty?)
          tmp = tmp.children.first
        end
        return PostorderIterator.new(self, tmp)
      end
    end
    
    def end_postorder
      PostorderIterator.new(self, @foot)
    end
    
    def insert(position, value)
      if(position.node.equal?(@root) && @root.equal?(@foot))
        @root = Node.new(value, nil, nil, @foot)
        @foot.set_older(@root);
        itetmp = position.clone
        itetmp.set_node(@root)
        position.set_node(@root)
        return itetmp
      elsif(position.node == nil &&  @root != nil)
        return nil
      elsif((self.equal?(position.tree)) && position.node != nil)
        pnode = position.node
        tmp = Node.new(value, pnode.parent, pnode.older, pnode.younger)
        if(tmp.parent != nil)
          pos1 = tmp.parent.children.index{|i| i.equal?(pnode) }
          tmp.parent.children[pos1] = tmp
          if(pos1 - 1 >= 0)
            tmp.parent.children[pos1 - 1].set_younger(tmp)
          end
          if(pos1 + 1 < tmp.parent.children.size)
            tmp.parent.children[pos1 + 1].set_older(tmp)
          end
        end
        tmp.append_child_node(pnode)
        pnode.set_parent(tmp)
        pnode.set_older(nil)
        pnode.set_younger(nil)
        if(@root.equal?(pnode))
          @root = tmp
          @foot.set_older(tmp)
        end
        itetmp = position.clone
        itetmp.set_node(tmp)
        return itetmp
      else
        return nil
      end
    end
    
    def append_child(position, value)
      if(position == nil || !(self.equal?(position.tree)))
        return nil
      end
      itetmp = position.clone
      itetmp.set_node(position.node.append_child(value))
      return itetmp
    end
    
    def append_child_tree(position, other_tree)
      if(position == nil || !(self.equal?(position.tree)) \
      || self.equal?(other_tree) || position.node == nil)
        return nil
      end
      position.node.append_child_node(other_tree.begin_preorder.node)
      return self
    end
    
    def erase(position)
      if(position == nil || !(self.equal?(position.tree)) \
      || position.node == nil || @foot.equal?(position.node))
        return nil
      end
      nodetmp = position.node
      if(nodetmp.older != nil)
        nodetmp.older.set_younger(nodetmp.younger)
      end
      if(nodetmp.younger != nil)
        nodetmp.younger.set_older(nodetmp.older)
      end
      if(nodetmp.parent != nil)
        nodetmp.parent.children.reject!{|v| v.equal?(nodetmp) }
      else
        @root = @foot
        @foot.set_older(nil)
      end
      position.set_node(nil)
      return nil
    end
    
    def output_dot(graphname = "G")
      strtmp = String.new
      strtmp << "digraph " << "\"" << graphname << "\" {\n"
      ite1 = self.begin_preorder
      end1 = self.end_preorder
      while(ite1 != end1)
        nodename1 = "N"
        nodename1 << ite1.node.object_id.to_s
        nodename1.gsub!(/-/){"M"}
        strtmp << "\t" << nodename1 << "[ label=\"" << ite1.get_value.to_s << "\" ];\n"
        ite1.node.children.each{ |child1|
          nodename2 = "N"
          nodename2 << child1.object_id.to_s
          nodename2.gsub!(/-/){"M"}
          strtmp << "\t" << nodename1 << " -> " << nodename2 << "\n"
        }
        ite1.get_next
      end
      strtmp << "}\n"
      return strtmp
    end
  end #---- class Tree
  
  #------------------------------------------------------------
  #    class TreeIteratorBase
  #------------------------------------------------------------
  class TreeIteratorBase
    def initialize(tree, node = nil)
      @tree = tree
      @node = node
      return self
    end
    
    def ==(other)
      if(self.class != other.class);
        return false
      else
        return @node.equal?(other.node)
      end
    end
    
    def set_node(node0)
      @node = node0
    end
    
    def get_value
      if(node == nil)
        return nil
      else
        return @node.value
      end
    end
    
    def set_value(val)
      if(node == nil)
        return nil
      else
        @node.set_value(val)
      end
    end
    attr_reader :node, :tree
  end
  
  #------------------------------------------------------------
  #    class PreorderIterator
  #------------------------------------------------------------
  class PreorderIterator < TreeIteratorBase
    def get_next
      @node = @node.preorder_increment
      return self
    end
    
    def get_prev
      @node = @node.preorder_decrement
      return self
    end
  end #---- class PreorderIterator
  
  #------------------------------------------------------------
  #    class PostorderIterator
  #------------------------------------------------------------
  class PostorderIterator < TreeIteratorBase
    def get_next
      @node = @node.postorder_increment
      return self
    end
    
    def get_prev
      @node = @node.postorder_decrement
      return self
    end
  end #---- class PostorderIterator
end #---- module Creek


